#include <stdio.h>
#include <time.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <unistd.h>

typedef long long unsigned int uint64;


extern int prime_factors(int n, int factors[]); 
extern int prime_factors_asm(int n, int factors[]);

int arrayEqual(int factors[], int factors1[], int nf);
int print_result(int factors[], int nf, int factors1[], int nf1, uint64 elapsed_time);
uint64 measure_tics_per_sec();

int main() {
  FILE *f = fopen("testcase", "r");
  uint64 total_time_c = 0;
  uint64 total_time_asm=0;
  uint64 ticks_per_sec = measure_tics_per_sec();

  // testing the first 10 numbers in "testcase"
  for (int itr = 0; itr < 20; itr++) {
    int n;                          // the number to be factored
    int nf;                         // number of factors
    int nf1;                        // number of factors given by the algorithm
    int factors[32];                // array of factors
    int factors1[32];               // array of factors given by the algorithm
    unsigned int lo,hi;             // used to measure   clock cycles
    uint64 ticks, start, end;       // used for mesuring clock cycles
    
    // read one row of the testcase
    fscanf(f,"%d %d", &n, &nf);
    for (int i = 0; i < nf; i++)
      fscanf(f,"%d", &factors[i]);
    
    printf(" C : %d", n);
    // measure clock cycles using the "rdtsc" assembly command 
    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    start = ((uint64)hi << 32) | lo;
        
    // run the algorithm
    nf1 = prime_factors(n,factors1);

    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    end = ((uint64)hi << 32) | lo;
    ticks = end-start;
    total_time_c += ticks;
    
    print_result(factors, nf, factors1, nf1, ticks);


    // zero out array
    for (int i = 0; i < 32; i++)
      factors1[i] = 0;
    

    printf("ASM: %d", n);
    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    start = ((uint64)hi << 32) | lo;
    
    // run the algorithm
    nf1 = prime_factors_asm(n,factors1);
    
    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    end = ((uint64)hi << 32) | lo;
    ticks = end-start;
    
    total_time_asm += ticks;

    print_result(factors, nf, factors1, nf1, ticks);

    putchar('\n');

    
    	   
  }
  fclose(f);

  printf("total_time (C) = %f\n", total_time_c / (double) ticks_per_sec);
  printf("total_time (ASM) = %f\n", total_time_asm/ (double) ticks_per_sec);


  
  
  return 0;
}


uint64 measure_tics_per_sec() {
  unsigned int lo,hi;             // used to measure   clock cycles
  uint64 start, end;       // used for mesuring clock cycles
    
  asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    start = ((uint64)hi << 32) | lo;
        
    sleep(1);

    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    end = ((uint64)hi << 32) | lo;


    return end-start;

}


int arrayEqual(int factors[], int factors1[], int nf) {
  if (nf == 0)
    return 1;
  return factors[nf-1] == factors1[nf-1] && arrayEqual(factors, factors1, nf-1);
}

int print_result(int factors[], int nf, int factors1[], int nf1, uint64 ticks) {

  printf("= %d", factors1[0]);
  for (int i = 1; i < nf1; i++)
    printf("*%d", factors1[i]);

  printf(" elapsed= %lld ticks ", ticks);

  
  int success = nf == nf1 && arrayEqual(factors, factors1, nf);
  if (success)
    printf(" (PASS)\n");
  else
    printf(" (FAIL!)\n");

  
  return success;
}
